/*
** toybox.c: container exploration toy
**
** This program implements a primitive "container"
** under the Linux 4.4 kernel. It requires an
** existing target directory (which will become
** the container's root directory), populated with
** (at a minimum) the bin directory and a bin/sh
** binary. toybox will create any other directories
** that it requires for it's operation.
**
** Usage: toybox [-h hostname] [-H] rootdir
** 
** Where: -h hostname sets the container hostname to specified value
**	  -H r/o bind mounts host /bin, /sbin, /lib, /lib64 and /usr,
**   and  rootdir is the name of the directory which toybox will use
**	    as it's own root directory
**
** toybox directory structure:
**
**   /
**   :
**   +- /bin		required, prepopulated, must include /bin/sh
**   :			(may be r/o bind mounted with host /bin)
**   :
**   +- /dev		constructed, toybox will bind mount host /dev to this mountpoint
**   :
**   +- /etc		optional, may hold prepopulated config files
**   :
**   +- /lib		optional, may be r/o bind mounted with host /lib
**   :
**   +- /lib64		optional, may be r/o bind mounted with host /lib64
**   :
**   +- /proc		constructed, toybox will mount procfs to this mountpoint
**   :
**   +- /sbin		optional, may hold prepopulated binaries
**   :			(may be r/o bind mounted with host /sbin)
**   :
**   +- /sys		constructed, toybox will mount sysfs to this mountpoint
**   :
**   +- /tmp		constructed, toybox will mount tmpfs to this mountpoint
**   :			(temporary mountpoint for host root filesystem)
**   :
**   +- /usr		optional, may hold prepopulated directories, binaries
**   :			(may be r/o bind mounted with host /usr)
**   :
**   +- /var		optional, may hold prepopulated data
**
**
** NB: compile with -D_GNU_SOURCE
** Note: compile with -DLIB64 to include /lib64 in ro bindmounts
**
** $Id: toybox.c,v 1.43 2023/09/25 13:10:04 lpitcher Exp lpitcher $
**
** Copyright (C) 2023  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
**
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>

/* pivot_root() function prototype not supplied by standard headers */
extern int pivot_root(const char *new_root, const char *put_old);

#define	CONTAINER	(CLONE_NEWUSER	|	\
			 CLONE_NEWNS	|	\
			 CLONE_NEWPID	|	\
			 CLONE_NEWUTS	|	\
			 CLONE_NEWNET	|	\
			 CLONE_NEWIPC)

/*
** NB:	CLONE_NEWUSER	gives us a new user namespace, and the process
**			gains CAP_SYS_ADMIN
**	CLONE_NEWNS	gives us a new mount namespace, which requires
**			the process have CAP_SYS_ADMIN to manipulate
**	CLONE_NEWPID	gives us a new PID namespace, which
**			gives access to the /proc filesystem,
**			and the special properties of PID 1
**			this namespace also requires the manipulating
**			process to have CAP_SYS_ADMIN
**	CLONE_NEWUTS	gives us a new UTS namespace, which requires
**			the process have CAP_SYS_ADMIN to manipulate
**	CLONE_NEWNET	gives us a new NETWORK namespace, which
**			gives us access to an isolated set of network
**			resources. Requires CAP_SYS_ADMIN
**	CLONE_NEWIPC	gives us a new IPC namespace, which
**			gives us access to an isolated set of Sys V IPC
**			and POSIX message queue resources. Requires
**			CAP_SYS_ADMIN
**	CLONE_NEWCGROUP	is not implemented in kernel 4.4
**
** To establish a paravirtual environment, an unprivileged process must use
** /at least/ CLONE_NEWUSER, CLONE_NEWNS, CLONE_NEWPID, and CLONE_NEWNET
*/

#define	FLAG_HOSTENV	(1<<0)

static int	toybox(const char *RootDir, const char *HostName, unsigned int toyflags);
static int 	assumeRoot(uid_t oldUid, gid_t oldGid);
static int 	isolateFs(const char *RootDir, unsigned int toyflags);
static int	idleProcess(const char *HostName, unsigned int toyflags);
static int	initProcess(const char *HostName, unsigned int toyflags);
static void 	sysMsg(const char *sev, int errnum, const char *msg, ...);

#define		ERROR(...) 	(sysMsg("Error",errno,__VA_ARGS__),0)
#define		WARNING(...)	sysMsg("Warning",errno,__VA_ARGS__)

#define		NENTRIES(x)	(sizeof(x)/sizeof(*x))

int main(int argc, char *argv[])
{
  int	success = 1,		/* assume success */
	opt;
  char	*HostName = NULL;
  unsigned int toyflags = 0;

  while ((opt = getopt(argc,argv,"h:H")) != -1)
  {
    switch (opt)
    {
      case 'h':
	free(HostName);
	HostName = strdup(optarg);
	break;

      case 'H':
	toyflags |= FLAG_HOSTENV;
	break;

      default:
	success = 0;
	break;
    }
  }

  if ((success) && (argc - optind == 1))
    success = toybox(argv[optind],HostName,toyflags);
  else
    fprintf(stderr,"Usage: %s [-h hostname] [-H] rootdir\n",argv[0]);

  return success?EXIT_SUCCESS:EXIT_FAILURE;
}

/*
** toybox(const char *RootDir, const char *HostName, unsigned int toyflags)
**
** Create and execute our toy container by unshare()ing
** its namespaces, changing its uid and gid, building its
** root filesystem and executing its idle and init processes.
**
** Accepts:	const char * pointer to root directory path string
**		const char * pointer to hostname string
**		an unsigned int containing various execution flags
**		  FLAG_HOSTENV will indicate whether (1) or not (0)
**		  to mount additional directories from the host
**		  filesystem
**
** Returns:	int 1 if successful in creating container
**		int 0 if container creation failed
*/
static int toybox(const char *RootDir, const char *HostName, unsigned int toyflags)
{
  int	success = 1;	/* assume success */
  uid_t	host_uid;
  gid_t	host_gid;

  umask(0);
  host_uid = geteuid();	/* must do before unshare() as uid changes within unshare */
  host_gid = getegid();	/* must do before unshare() as gid changes within unshare */

  /* change our namespaces */
  if (unshare(CONTAINER))
    success = ERROR("unshare()");

  /* change our uid, gid to 0,0 */
  if (success)
    success = assumeRoot(host_uid,host_gid);

  /* set up our isolated root directory */
  if (success)
    success = isolateFs(RootDir,toyflags);

  /* run our idle and init processes within our isolated root filesystem */
  if (success)
    success = idleProcess(HostName,toyflags);

  return success;
}

/*
** assumeRoot(oldUid, oldGid)
**
** Update the process entry in /proc to map our 
** outside-the-box uid and gid to the inside-the-box
** root user uid and gid, matching the effects
** of our new CLONE_NEWUSER capabilities.
**
** Accepts:	uid_t oldUid - our outside-the-box UID,
**		gid_t oldGid - our outside-the-box GID
**
** Returns: 	int 1  - success
**	    	int 0  - failure
*/
static int assumeRoot(uid_t oldUid, gid_t oldGid)
{
  int	success = 1,	/* assume success */ 
	fd;
  FILE *map;

  /*
  ** Map our uid into the container's root uid (UID 0)
  */
  if (map = fopen("/proc/self/uid_map","w"))
  {
    fprintf(map,"0 %lu 1\n",(unsigned long)oldUid);
    fclose(map);
  }
  else success = ERROR("fopen(\"/proc/self/uid_map\")");

  /*
  ** Some kernels require /proc/self/setgroups set to "deny"
  ** before we can remap our GID
  */
  if (success  && ((fd = open("/proc/self/setgroups",O_WRONLY)) >= 0))
  {
    write(fd,"deny",4);
    close(fd);
  }

  if (success)
  {
    /*
    ** Map our gid into the container's root gid (GID 0)
    */
    if (map = fopen("/proc/self/gid_map","w"))
    {
      fprintf(map,"0 %lu 1\n",(unsigned long)oldGid);
      fclose(map);
    }
    else success = ERROR("fopen(\"/proc/self/uid_map\")");
  }

  return success;
}

/*
** isolateFs(const char *RootDir, unsigned int toyflags)
**
** Build our container's root filesystem by creating
** required subdirectories in target rootdir, then
** mount()ing various requisite resources on those
** subdirectories.
**
** Accepts:	a const char pointer to the pathname of the root filesystem
**		an unsigned int containing various execution flags
**		  FLAG_HOSTENV will indicate whether (1) or not (0)
**		  to mount additional directories from the host
**		  filesystem
**
** Returns:	int 1 if successful creation of container
**		 root filesystem hierarchy
**		int 0 if container root filesystem cannot
**		 be properly built
*/
static int isolateFs(const char *RootDir, unsigned int toyflags)
{
  int	success = 1;	/* assume success */


  /* CHECKPOINT:
  **   First, we will isolate our container's view of the host filesystem,
  **   and relocate our CWD to the root of the intended container filesystem
  */

  /* Ensure that none of our actions here affect host mountspace and
  ** no actions in the host mountspace affect our mounts.
  ** This satisfies a condition of pivot_root(2), which requires that
  ** the propagation type of the parent mount of the current root
  ** directory must not be MS_SHARED.
  **
  **	MS_PRIVATE makes this mount "private"; mount and unmount events
  **	  do not propogate into or out of this mount.
  **	MS_REC will recursively change the propagation type of all of
  **	  the mounts in a subtree
  ** source, filesystemtype and data arguments are ignored.
  */
  if (mount(NULL,"/",NULL,MS_REC|MS_PRIVATE,NULL))
    success = ERROR("mount(,\"/\",,)");

  /* make our root directory into a mountpoint in prep for pivot_root
  ** This satisfies a condition of pivot_root(2), which requires that
  ** the "new root" not be on the same filesystem as the current root
  ** by making our "new root" directory into a filesystem of it's own.
  **
  ** 	MS_BIND performs a bind mount.  A bind mount makes a file or a
  **	  directory subtree visible at another point within the
  **	  single directory hierarchy. 
  **	MS_NOSUID ensures that the mount does not honor set-user-ID and
  **	  set-group-ID bits or file capabilities when executing programs
  **	  from this filesystem.
  ** filesystemtype and data arguments are ignored.
  **
  ** Note: at this point, our process CANNOT use our target root directory
  ** as its CWD. To do so will cause pivot_root() to fail. 
  */
  if (success && mount(RootDir,RootDir,NULL,MS_BIND|MS_NOSUID,NULL))
    success = ERROR("mount(\"%s\",\"%s\",NULL,MS_BIND|MS_NOSUID,NULL)",RootDir,RootDir);

  /* Continue operations from within our intended root directory */
  if (success && chdir(RootDir))
    success = ERROR("chdir(\"%s\")",RootDir);


  /* CHECKPOINT:
  **   At this point, we have isolated our container's view of the host filesystem,
  **   and relocated our CWD to the root of the intended container filesystem
  **   Now, we prepare and populate our container filesystem's directory structure
  */

  /* Remove and recreate all of the REQUISITE unpopulated subdirectories */
  if (success)
  {
    struct {
      char *path;
      mode_t mode;
    } directory[] = {	{"dev",0755},		/* mount point for /dev bind mount */
			{"proc",0755},		/* mount point for proc mount */
			{"sys",0751},		/* mount point for sysfs mount */
			{"tmp",01777}		/* mount point for pivot_root old_root, tmpfs */
		    };

    /* delete existing local directories (in reverse order, in case of dependencies) */
    for (int index = NENTRIES(directory); index > 0; --index)
      if (unlink(directory[index-1].path) && rmdir(directory[index-1].path))
	WARNING("rmdir(\"%s\")",directory[index-1].path);

    /* create new local directories (in list order) */
    for (int index = 0;(success) && (index < NENTRIES(directory)); ++index)
      if (mkdir(directory[index].path,directory[index].mode))
	success = ERROR("mkdir(\"%s\",%04o)",directory[index].path,(unsigned int)directory[index].mode);
  }


  /* CHECKPOINT:
  **   At this point, we have ensured that the container's (intended root) directory
  **   has the directory structure to facilitate all of our REQUISITE BIND mounts
  **   Now, we go ahead and BIND MOUNT host directories to the corresponding container
  **   mountpoints
  */

  /* Populate our /dev directory with direct link to host /dev
  ** 	MS_BIND performs a bind mount.  A bind mount makes a file or a
  **	  directory subtree visible at another point within the
  **	  single directory hierarchy.
  **    MS_REC makes this a recursive bind mount where all submounts
  **	  under the source subtree (other than unbindable mounts) are
  **	  also bind mounted at the corresponding location in the target
  **	  subtree.
  ** filesystemtype and data arguments are ignored.
  */
  if (success && mount("/dev","dev",NULL,MS_BIND|MS_REC,NULL))
    success = ERROR("mount(\"dev\")");


  /* CHECKPOINT:
  **   At this point, we have completed the set up of the REQUISITE portions
  **   of our new filesystem, and now will address the OPTIONAL portions before
  **   we commit our container to using the new filesystem exclusively.
  */

  /* OPTIONALLY ReadOnly Recursive BIND mount the host's program and library directories */
  if (toyflags & FLAG_HOSTENV)
  {
    /* 
    ** Note: here, we have a list of host directories (and their corresponding container directories)
    ** that we wish to bind mount in order to access the binaries and libraries within. We will
    ** mount these as recursive bind mounts, set to read-only, so that we can access (but not
    ** alter) the contents.
    */
    struct {
      char *source;
      char *target;
    } optHostDir[] = {	{"/bin",   "bin"},
			{"/lib",   "lib"},
#ifdef LIB64
			{"/lib64", "lib64"},	/* compile time optional for 64bit systems */
#endif
			{"/sbin",  "sbin"},
			{"/usr",   "usr"}
		   };

    /*
    ** Note: we perform this "R/O Recursive Bind mount" in two stages:
    ** 1) recursive bind mount the host source directory to our target directory, and then
    ** 2) remount the target directory, with the read-only flag set
    */

    for (int index = 0; (success) && (index < NENTRIES(optHostDir)); ++index)
    {
      unlink(optHostDir[index].target);		/* get rid of regular file with target name */
      mkdir(optHostDir[index].target,0);	/* create directory with target name, no permissions */

      if (mount(optHostDir[index].source,optHostDir[index].target,NULL,MS_BIND|MS_REC,NULL))
	success = ERROR("mount(\"%s\",\"%s\",NULL,MSBIND_MS_REC,NULL)",
			optHostDir[index].source,optHostDir[index].target);
      else
	if (mount(optHostDir[index].source,optHostDir[index].target,NULL,MS_REMOUNT|MS_BIND|MS_REC|MS_RDONLY,NULL))
	  WARNING("mount(\"%s\",\"%s\",MS_REMOUNT|MS_BIND|MS_REC|MS_RDONLY,NULL)",
		   optHostDir[index].source,optHostDir[index].target);
    }
  }


  /* CHECKPOINT:
  **   NOW, we commit our container to using the new filesystem exclusively
  **    1) pivot_root()	to reroot the directory structure
  **    2) chdir()	to ensure that our cwd resides in the new directory structure
  */

  /* Isolate our directory space and occupy the toy box
  ** Note: The following restrictions apply to new_root and put_old:
  **	   - They must be directories.
  **	   - new_root and put_old must not be on the same filesystem
  **	     as the current root.
  **	   - put_old must be underneath new_root, that is, adding a
  **	     nonzero number of /.. to the string pointed to by put_old
  **	     must yield the same directory as new_root.
  **	   - new_root must be a path to a mount point, but can't be "/".
  **	     A path that is not already a mount point can be converted
  **	     into one by bind mounting the path onto itself.
  **	   - The propagation type of the parent mount of new_root and
  **	     the parent mount of the current root directory must not be
  **	     MS_SHARED; similarly, if put_old is an existing mount point,
  **	     its propagation type must not be MS_SHARED. These
  **	     restrictions ensure that pivot_root() never propagates any
  **	     changes to another mount namespace.
  **	   - No other filesystem may be mounted on put_old.
  */
  if (success && pivot_root(".","tmp"))
    success = ERROR("pivot_root(\".\",\"tmp\")");

  /* pivot_root(2) recommends that we chdir("/") to ensure correct cwd */
  if (success && chdir("/"))
    success = ERROR("chdir(\"/\")");


  return success;
}

/*
** idleProcess(toyParms_t toyParms, unsigned int toyflags)
**
** Execute container's "idle" (PID 0) 
** and "init" (PID 1) processes
**
** Accepts:	a const char pointer to new hostname, or NULL
**		an unsigned int containing various execution flags
**		  FLAG_HOSTENV will indicate whether (1) or not (0)
**		  to mount additional directories from the host
**		  filesystem
**
** Returns:	int 1 if successful execution of container
**		int 0 if failure of init process
*/
static int idleProcess(const char *HostName, unsigned int toyflags)
{
  int	success = 1,		/* assume success */
	child_status;
  pid_t	child;

  switch (child = fork())
  {
    case -1:	/* still in parent: fork() failed */
      success = ERROR("fork()");
      break;

    case 0:	/* in child, which takes the init process role */
      success = initProcess(HostName,toyflags);
      break;

    default:	/* in parent, which takes on the idle process role */
      wait(&child_status);
      if (!WIFEXITED(child_status) || (WEXITSTATUS(child_status) != 0))
	success = 0;

      /* unmount our private copies of mounted filesystems */
      umount2("/tmp",MNT_DETACH);	/* not strictly necessary */
      umount2("/dev",MNT_DETACH);	/* not strictly necessary */
      umount2("/sys",MNT_DETACH);	/* not strictly necessary */
      umount2("/proc",MNT_DETACH);	/* not strictly necessary */
  }

  return success;
}

/*
** initProcess(const char *HostName, unsigned int toyflags)
**
** Execute the container's init (PID 1) process
**
** Accepts:	a const char pointer to new hostname, or NULL
**		an unsigned int containing various execution flags
**		  FLAG_HOSTENV will indicate whether (1) or not (0)
**		  to mount additional directories from the host
**		  filesystem
**
** Returns:	int 1 if successful execution of container
**		int 0 if failure of init process
**
** Note:	This function will generate warning messages
**		 under certain circumstances. These informational
**		 messages can be ignored.
*/
static int initProcess(const char *HostName, unsigned int toyflags)
{
  int	success = 1;		/* assume success */


  /* Populate our /proc directory */
  if (mount("none","/proc","proc",MS_REC,NULL))
    success = ERROR("mount(\"/proc\")");

  /* Mount a true sysfs to our (isolated) /sys directory */
  if (success && mount("none","/sys","sysfs",0,NULL))
    success = ERROR("mount(\"/sys\")");

  /* Discard our tie to the host root directory */
  if (success && umount2("/tmp",MNT_DETACH))
    success = ERROR("umount2(\"/tmp\")");

  /* Give ourselves a tempfs for our /tmp directory */
  if (success && mount("none","tmp","tmpfs",
			MS_NOSUID|MS_NODEV|MS_NOEXEC|MS_NOATIME,
			"size=16m,nr_inodes=4k,mode=01777"))
    success = ERROR("mount(\"tmp\")");

  /* Become session leader, process group leader, with no controlling terminal */
  if (setsid() == -1)	success = ERROR("setsid()");

  /* If required, set our container's UTS hostname */
  if (success && HostName)
    if (sethostname(HostName,strlen(HostName)))
      WARNING("sethostname(\"%s\",%lu)",HostName,(unsigned long)strlen(HostName));

  /* Run a very simple init with a single child process running shell */
  if (success)
  {
    pid_t	child;
    char	*envp[] = {"PATH=/sbin:/bin:/usr/sbin:/usr/bin:",NULL};

    switch (child = fork())	/* VERY SIMPLE INIT */
    {
      case 0:	/* this "only child" runs a shell */
	execle("/bin/sh","-sh",NULL,envp);
	ERROR("execle(\"/bin/sh\")");
	_exit(127);
	break;

      case -1:	/* fork() failed - so we quit */
	success = ERROR("fork()");
	break;

      default:	/* parent reaps child and orphan status values */
	while (wait(NULL) != child)
	  continue;
	break;
    }
  }

  return success;
}

/*
********************************** REPORTING UTILITIES **********************************
*/

/*
** sysMsg(const char *sev, int errnum, const char *msg, ...)
**
** Report a syscall error
**
** Accepts:	a const char pointer to a severity string,
**		an integer, derived from the errno returned by the syscall,
**		a const char pointer to a message format string, and
**		optional arguments as required to satisfy the message format string
**
** Returns:	nothing
*/
static void sysMsg(const char *sev, int errnum, const char *msg, ...)
{
  va_list ap, pp;
  char	*buf;
  int	buflen;

  va_start(ap,msg);
  va_copy(pp,ap);
  if ((buflen = vsnprintf(NULL, 0, msg, pp)) >= 0)
  {
    if (buf = malloc(buflen += 2))
    {
      va_copy(pp,ap);
      vsnprintf(buf,buflen,msg,pp);
    }
  }
  else buf = NULL;
  va_end(ap);

  fprintf(stderr,"%s: %s returned \"%s\"\n",sev,buf?buf:"a system call",strerror(errnum));
  free(buf);
}

/*
************************************** CHANGE LOG **************************************
**
** $Log: toybox.c,v $
** Revision 1.43  2023/09/25 13:10:04  lpitcher
** comment cleanup
** mount tmpfs with mode=01777
**
** Revision 1.42  2023/09/25 04:21:05  lpitcher
** removed transitional /proc and /sys bind mounts and their cleanup
** pivot_root() moves old_root to new /tmp
** move umount(old_root) from isolateFs to initProcess()
**
** Revision 1.41  2023/09/23 09:12:52  lpitcher
** use /tmp instead of /mnt for temporary proc/sys/root mounts
** moved /tmp tmpfs mount from isolateFs() to initProcess()
**
** Revision 1.40  2023/09/23 04:01:11  lpitcher
** expand comments around initial propogation mount of host root
** make "source" NULL on initial host root propogation mount.
** expand comments around bind mount of target directory.
** expand comments around bind mount of target dev directory.
** expand comments around bind mount of temporary proc directory.
** expand comments around bind mount of temporary sys directory.
** expand comments around pivot_root() call
**
** Revision 1.39  2023/09/17 20:59:28  lpitcher
** ensure that process 1 is session leader, process group leader
** without controlling terminal
**
** Revision 1.38  2023/09/12 17:46:40  lpitcher
** implement a very simple init which launches our root shell
** then waits on child and orphan terminations
**
** Revision 1.37  2023/08/25 15:18:43  lpitcher
** refactored sysMsg() message length logic
** changed sysMsg() return value to void
** corrected rmdir() error msg to properly indicate octal mode
** minor comment cleanups
**
** Revision 1.36  2023/08/25 14:49:15  lpitcher
** more comments around isolateFs() directory management
** merge readonly bind mount logic back into isolateFs()
** pass toyflags to idleProcess(), initProcess() in prep
**  for mount cleanup (via FLAG_HOSTENV) and choice between
**  init(1) and sh(1) as container init system.
**
** Revision 1.35  2023/08/25 00:58:12  lpitcher
** streamlined some directory creation logic
** minor comment changes
**
** Revision 1.34  2023/08/24 22:57:29  lpitcher
** redesigned error, warning reporting
** changed all uses of SYSERROR to new error, warning reporting
**
** Revision 1.33  2023/08/24 20:02:01  lpitcher
** corrected "r/o bind mount" logic
**
** Revision 1.32  2023/08/21 19:48:42  lpitcher
** generalize execution flags
**
** Revision 1.31  2023/08/21 17:51:31  lpitcher
** conditional compilation around /lib64 bind mount
** #define LIB64 to enable
**
** Revision 1.30  2023/08/21 17:48:56  lpitcher
** more comment cleanup
**
** Revision 1.29  2023/08/21 17:45:38  lpitcher
** comment cleanup
**
** Revision 1.28  2023/08/21 17:40:10  lpitcher
** additional comments re r/o bind mounts
**
** Revision 1.27  2023/08/21 17:34:00  lpitcher
** added -H flag to demand bind mount of additional host dirs
** on demand, r/o bind mount host /bin, /sbin, /lib, /lib64
** and /usr to corresponding toybox mountpoints.
**
** Revision 1.26  2023/08/11 15:07:44  lpitcher
** root shell invoked as "login" shell
**
** Revision 1.25  2023/08/10 15:13:43  lpitcher
** removed tParameters. simplified all parameters to string arguments
** simplified final process argument success test, arg count test
** use stdio to update proc uid_map and gid_map
**
** Revision 1.24  2023/07/11 01:06:15  lpitcher
** minor comment cleanup
**
** Revision 1.23  2023/05/05 18:27:20  lpitcher
** added GPL v2 licence
**
** Revision 1.22  2023/01/27 01:59:47  lpitcher
** reverted code that creates own dev entries.
** will continue with valid code until we determine
** what caused the toydev failures
** This version logically identical to 1.20
**
** Revision 1.20  2023/01/22 14:30:34  lpitcher
** pass name of root directory in toyopts_t toyOptions structure
** remove rootdir reference from function declarations and definitions
** in recognition of expanded function of tOptions control block,
**   renamed tOptions to tParameters
**   renamed toyOpts_t to toyParms_t
**   renamed toyOpts to toyParms
**   renamed toyOptions to toyParameters
**   renamed FLAG_HOSTNAME to PARM_HOSTNAME
**
** Revision 1.19  2023/01/20 21:38:15  lpitcher
** after pivot_root(), chdir("/") to ensure correct cwd
** at idle termination, unmount our private copies of mounted filesystems
** pass toyopts_t toyParameters by value to subordinate functions
**
** Revision 1.18  2023/01/16 00:03:48  lpitcher
** added more commentary around directory manipulation
** isolate and mount our own sysfs to /sys
**
** Revision 1.17  2023/01/09 16:05:08  lpitcher
** use absolute paths once we pivot_root
**
** Revision 1.16  2023/01/09 15:44:40  lpitcher
** final cleanup of unused vars - fixed some unintended compile errors
**
** Revision 1.15  2023/01/09 15:43:44  lpitcher
** clean up unused vars
**
** Revision 1.14  2023/01/09 15:42:20  lpitcher
** simplified all isolateFs() tests
**
** Revision 1.13  2023/01/09 15:35:43  lpitcher
** simplified all initProcess() tests
**
** Revision 1.12  2023/01/09 15:31:37  lpitcher
** simplified sethostname call
**
** Revision 1.11  2023/01/09 15:29:30  lpitcher
** adjusted init PATH
**
** Revision 1.10  2023/01/09 13:31:46  lpitcher
** expanded container creation commentary
**
** Revision 1.9  2023/01/09 13:21:16  lpitcher
** replaced flags and hostname with toyParms_t structure
** set flag in toyParms_t for hostname when given
** use flag in toyParms_t to determine hostname when needed
**
** Revision 1.8  2023/01/09 07:29:03  lpitcher
** allow empty hostname to reset UTS hostname
**
** Revision 1.7  2023/01/09 07:00:21  lpitcher
** correct spelling errors
**
** Revision 1.6  2023/01/09 06:42:23  lpitcher
** propogate flags variable to assumeRoot(), isolateFs(), idleProcess(), and initProcess()
** global rename status variable to success, to better reflect its purpose
**
** Revision 1.5  2023/01/09 06:02:47  lpitcher
** refactored toybox() logic
** added parens to ERROR macro
**
** Revision 1.4  2023/01/09 05:43:19  lpitcher
** added empty hostname check to initProcess
** added descriptive comments for each function
**
** Revision 1.3  2023/01/09 05:30:53  lpitcher
** added sethostname() to initProcess()
** added hostname parameter to idleProcess(), initProcess()
** pass hostname to idleProcess, initProcess
**
** Revision 1.2  2023/01/09 05:18:46  lpitcher
** added RCS log
**
*/
